
#include "mqtt/MQTTClient.h"
#include "mqtt/MQTTConnect.h"
#include "wrapper/MQTTOsWrapper.h"
#include "em_chip.h"
#include "em_rtc.h"
#include "rtcdriver.h"

#define MQTT_PUB_TOPIC              "$aws/things/display/shadow/update"
#define MQTT_SUB_TOPIC              "$aws/things/display/shadow/update/accepted"
#define MQTT_CLIENT_ID              "Zentri"
#define MQTT_USER                   ""
#define MQTT_PASSWORD               ""
#define MQTT_HOST                   "A1NNFAKJYT1OHL.iot.us-east-1.amazonaws.com"
#define MQTT_PORT                   MQTT_SECURE_PORT

#define MQTT_PUBLISH_PERIOD         5000
#define MQTT_MESSAGE                "{\"state\":{\"desired\":{\"message\":\"Time now: %lu\"}}}"
#define COMMAND_START_PERIODIC      "{\"state\":{\"desired\":{\"message\":\"PERIODIC\"}}"
#define COMMAND_STOP_PERIODIC       "{\"state\":{\"desired\":{\"message\":\"STOP\"}}"
#define COMMAND_READ_ONCE           "{\"state\":{\"desired\":{\"message\":\"READ\"}}"

static RTCDRV_TimerID_t periodic_timer;
static MQTTClient client = DefaultClient;
static bool publish_now = false;

static void publish();
static void timer_handler( RTCDRV_TimerID_t id, void *user);
static void receive_handler(MessageData* rx_msg);

/******************************************************************************
 * @brief  Main
 *****************************************************************************/
int main(void)
{
    int ret;
    unsigned char send_buf[512];
    unsigned char recv_buf[2048];

    Network network;
    MQTTPacket_connectData data = MQTTPacket_connectData_initializer;

    /* Chip errata */
    CHIP_Init();

    /* Initialize RTC timer. */
    RTCDRV_Init();
    RTCDRV_AllocateTimer(&periodic_timer);

    data.MQTTVersion = 4;
    data.clientID.cstring = MQTT_CLIENT_ID;
    data.username.cstring = MQTT_USER;
    data.password.cstring = MQTT_PASSWORD;

    NetworkInit(&network);
    RTCDRV_Delay(250);

    ret = NetworkConnect(&network, MQTT_HOST, MQTT_PORT);
    if(ret == FAILURE)
    {
        // error, return
        return ret;
    }
    RTCDRV_Delay(250);

    MQTTClientInit(&client, &network, 10000, send_buf, sizeof(send_buf), recv_buf, sizeof(recv_buf));
    RTCDRV_Delay(250);

    ret = MQTTConnect(&client, &data);
    if(ret != SUCCESS)
    {
        // error, return
        return ret;
    }
    RTCDRV_Delay(250);


    MQTTSubscribe(&client, MQTT_SUB_TOPIC, QOS0, receive_handler);
    RTCDRV_Delay(250);

    // running forever to receive messages from server
    while(true)
    {
        // can not execute UART blocking operations in timer handler,
        // so, we just set a flag and check it here in main app..
        if(publish_now == true)
        {
            publish_now = false;
            publish();
        }

        MQTTYield(&client, 500);
    }

    // won't reach here as MQTTRun is blocking
    MQTTUnsubscribe(&client, MQTT_SUB_TOPIC);
    MQTTDisconnect(&client);
    NetworkDisconnect(&network);

    return 0;
}

/******************************************************************************
 * @brief  Static functions
 *****************************************************************************/

static void publish()
{
    char buf[100];
    MQTTMessage message;

    sprintf(buf, MQTT_MESSAGE, (long unsigned int) TICKS_TO_MSEC(RTCC_CounterGet())/1000);
    message.qos = QOS0;
    message.retained = false;
    message.dup = false;
    message.payload = (void*)buf;
    message.payloadlen = strlen(buf);

    MQTTPublish(&client, MQTT_PUB_TOPIC, &message);
}

/*****************************************************************************/
static void timer_handler( RTCDRV_TimerID_t id, void *user)
{
    // time to publish.. set the flag and let the main app executes the publish
    publish_now = true;
}

/*****************************************************************************/
static void receive_handler(MessageData* rx_msg)
{
    char cmd[2048];

    snprintf(cmd, rx_msg->message->payloadlen + 1, rx_msg->message->payload);
    if(strncmp(cmd, COMMAND_READ_ONCE, strlen(COMMAND_READ_ONCE)) == 0)
    {
        publish_now = true;
    }
    else if(strncmp(cmd, COMMAND_START_PERIODIC, strlen(COMMAND_START_PERIODIC)) == 0)
    {
        RTCDRV_StartTimer(periodic_timer, rtcdrvTimerTypePeriodic, MQTT_PUBLISH_PERIOD, timer_handler, NULL);
        publish_now = true;
    }
    else if(strncmp(cmd, COMMAND_STOP_PERIODIC, strlen(COMMAND_STOP_PERIODIC)) == 0)
    {
        RTCDRV_StopTimer(periodic_timer);
    }
}
